from .base import EventDataSerializer
from .metrica import MetricaEventsJsonSerializer
from .opta import OptaSerializer
from .sportec import SportecEventSerializer

# from .metrica import MetricaEventSerializer
from .statsbomb import StatsBombSerializer
from .wyscout import WyscoutSerializer
